# SPDX-FileCopyrightText: 2025 Blender Foundation
#
# SPDX-License-Identifier: GPL-3.0-or-later


import bpy
from bpy.props import (
    PointerProperty,
    StringProperty,
    IntProperty,
    BoolProperty,
)
from bpy.types import AddonPreferences

import pathlib


FILEFORMATS_TO_EXTENSIONS = {
    'BMP': '.bmp',
    'IRIS': '.sgi',
    'PNG': '.png',
    'JPEG': '.jpeg',
    'JPEG2000': '.jp2',
    'TARGA': '.tga',
    'TARGA_RAW': '.tga',
    'CINEON': '.cin',
    'DPX': '.dpx',
    'OPEN_EXR_MULTILAYER': '.exr',
    'OPEN_EXR': '.exr',
    'HDR': '.hdr',
    'TIFF': '.tif',
    'WEBP': '.webp',
    'FFMPEG': "",
}

def texture_autosave_filepath_gen(im, blendfile_path, do_overwrite):
    prefs = bpy.context.preferences.addons[__name__].preferences
    # If image has a file path, and overwrite is enabled, 'save' the image.
    im_filepath_abs = pathlib.Path(im.filepath_from_user())
    if do_overwrite and im_filepath_abs:
        return im_filepath_abs, False
    # Define autosave root directory and ensure it exists.
    root_dir = pathlib.Path(bpy.app.tempdir)
    if not prefs.use_temp_dir and blendfile_path:
        root_dir = blendfile_path.parent
    autosave_dir = pathlib.Path(bpy.path.abspath(prefs.dirpath, start=root_dir))
    autosave_dir.mkdir(parents=True, exist_ok=True)
    # Define image filename.
    im_filename_base = im_filepath_abs.stem if im_filepath_abs else im.name
    im_filename_prefix = (pathlib.Path(bpy.data.filepath).stem if bpy.data.filepath
                          else pathlib.Path(bpy.app.tempdir).name)
    im_filename_suffix = "autosave"
    im_filename_extension = im_filepath_abs.suffix if im_filepath_abs else FILEFORMATS_TO_EXTENSIONS[im.file_format]
    im_filename = f"{im_filename_prefix}_{im_filename_base}_{im_filename_suffix}{im_filename_extension}"
    # Return final filepath.
    return autosave_dir / im_filename, True


def texture_autosave_process(blendfile_path, do_overwrite):
    blendfile_path = pathlib.Path(blendfile_path)
    for im in bpy.data.images:
        # Packed images are assumed saved together with the blendfile.
        if not im.is_dirty:
            continue
        if not im.has_data:
            continue
        if im.file_format in {'FFMPEG'}:
            continue
        if im.packed_file or im.packed_files:
            continue
        im_filepath, save_copy = texture_autosave_filepath_gen(im, blendfile_path, do_overwrite)
        print(f"SAVING '{im.name}' image at '{im_filepath}' {'(as copy)' if save_copy else '(overwrite)'}")
        im.save(filepath=str(im_filepath), save_copy=save_copy)


def textures_autosave_timer():
    print("TEXTURE AUTOSAVE TIMER")
    prefs = bpy.context.preferences.addons[__name__].preferences
    texture_autosave_process(bpy.data.filepath, prefs.do_overwrite_on_timer)
    # auto_save_time is in minutes.
    return bpy.context.preferences.filepaths.auto_save_time * 60


def textures_autosave_timer_update(preferences, context):
    if preferences.use_timer:
        if not bpy.app.timers.is_registered(textures_autosave_timer):
            bpy.app.timers.register(textures_autosave_timer, persistent=True)
    else:
        if bpy.app.timers.is_registered(textures_autosave_timer):
            bpy.app.timers.unregister(textures_autosave_timer)


@bpy.app.handlers.persistent
def textures_autosave_handler(blendfile_path):
    print("TEXTURE AUTOSAVE HANDLER", blendfile_path)
    prefs = bpy.context.preferences.addons[__name__].preferences
    texture_autosave_process(blendfile_path, prefs.do_overwrite_on_save)


def textures_autosave_handler_update(preferences, context):
    if preferences.use_save_handler:
        if textures_autosave_handler not in bpy.app.handlers.save_pre:
            bpy.app.handlers.save_pre.append(textures_autosave_handler)
    else:
        if textures_autosave_handler in bpy.app.handlers.save_pre:
            bpy.app.handlers.save_pre.remove(textures_autosave_handler)


class TexturesAutosavePreferences(AddonPreferences):
    # This must match the add-on name, use `__package__`
    # when defining this for add-on extensions or a sub-module of a python package.
    bl_idname = __name__

    use_timer: BoolProperty(
        name="Use Autosave Timer",
        description="Use a timer (with same interval as for blendfile itself) to save modified images",
        default=True,
        update=textures_autosave_timer_update,
    )
    do_overwrite_on_timer: BoolProperty(
        name="Overwrite Existing on Autosave",
        description="When the modified image already exists on disk, overwrite the existing file. "
                    "If False, only files in 'Autosave Directory' will be overwritten",
        default=False,
    )
    use_save_handler: BoolProperty(
        name="Save On Blendfile Saving",
        description="Also write images when manually saving the blendfile",
        default=True,
        update=textures_autosave_handler_update,
    )
    do_overwrite_on_save: BoolProperty(
        name="Overwrite Existing on Blendfile Saving",
        description="When the modified image already exists on disk, overwrite the existing file. "
                    "If False, only files in 'Autosave Directory' will be overwritten",
        default=True,
    )
    use_temp_dir: BoolProperty(
        name="Use Temp Folder as Root",
        description="Always use Blender's temp directory as root for 'Autosave Directory' to write auto-saved images, "
                    "instead of the written blendfile's directory "
                    "(NOTE: temp directory is always the root when current blendfile has not yet been saved)",
        default=False,
    )
    dirpath: StringProperty(
        name="Autosave Directory",
        description="Where to auto-save images when not overwriting existing files, "
                    "start with '//' to use a path relative to the saved blendfile (or the temp folder if enabled)",
        subtype='DIR_PATH',
        default="//autosave_images",
    )

    def draw(self, context):
        layout = self.layout
        layout.label(text="WARNING: Only handles static images currently, "
                          "images sequences or videos may not be properly auto-saved")
        row = layout.row()
        row.prop(self, "use_timer")
        sub = row.row()
        sub.enabled = self.use_timer
        sub.prop(self, "do_overwrite_on_timer")
        row = layout.row()
        row.prop(self, "use_save_handler")
        sub = row.row()
        sub.enabled = self.use_save_handler
        sub.prop(self, "do_overwrite_on_save")
        row = layout.row()
        row.prop(self, "dirpath")
        row.prop(self, "use_temp_dir")


classes = [TexturesAutosavePreferences]


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    textures_autosave_handler_update(bpy.context.preferences.addons[__name__].preferences, bpy.context)
    textures_autosave_timer_update(bpy.context.preferences.addons[__name__].preferences, bpy.context)


def unregister():
    if bpy.app.timers.is_registered(textures_autosave_timer):
        bpy.app.timers.unregister(textures_autosave_timer)

    if textures_autosave_handler in bpy.app.handlers.save_pre:
        bpy.app.handlers.save_pre.remove(textures_autosave_handler)

    for cls in classes:
        bpy.utils.unregister_class(cls)


if __name__ == "__main__":
    register()
